#######################################################################
#                            Flux methods                             #
#######################################################################



@with_signature function flux(equation::EulerEquation)
  @accepts State(rho, q, E)
  @unpack eos = equation
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  flx_rho = q
  flx_q   = q^2 / rho + p
  flx_E   = q / rho * (E + p)
  @returns flx_rho, flx_q, flx_E
end



@with_signature function speed(equation::EulerEquation)
  @accepts State(rho, q, E)
  @unpack eos = equation
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  v = abs(q / p) + sqrt(abs(eos.gamma * p / rho))
  @returns v
end


@with_signature function internal_energy(::EulerEquation)
  @accepts rho, q, E
  eps = E / rho - q^2 / (2 * rho^2)
  @returns eps
end


@with_signature function pressure(equation::EulerEquation{EquationOfState.IdealGas})
  @accepts rho, eps
  @unpack eos = equation
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  @returns p
end


@with_signature function entropy(equation::EulerEquation{EquationOfState.IdealGas})
  @accepts rho, q, p
  @unpack eos  = equation
  gamma = eos.gamma
  S = rho / (gamma - 1) * log(p / rho^gamma)
  @returns S
end


@with_signature function entropy_flux(equation::EulerEquation)
  @accepts rho, q, S
  @unpack eos  = equation
  flx_S = q / rho * S
  @returns flx_S
end



#######################################################################
#                                 2D                                  #
#######################################################################


LLF(nfl,nfr,ul,ur,v) = 1/2 * (nfl + nfr - v * (ur - ul))


@with_signature function internal_energy(::EulerEquation2d)
  @accepts rho, q, E
  eps = E / rho - q^2 / (2 * rho^2)
  @returns eps
end


@with_signature function pressure(equation::EulerEquation2d)
  @accepts rho, qx, qy, E
  q = sqrt(qx^2+qy^2)
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  @returns p
end


@with_signature function flux(equation::EulerEquation2d)
  @accepts rho, qx, qy, E, p
  @unpack eos = equation
  q = sqrt(qx^2+qy^2)
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  flx_rho_x = qx
  flx_rho_y = qy
  flx_qx_x = qx^2/rho + p
  flx_qx_y = qx*qy / rho
  flx_qy_y = qy^2/rho + p
  flx_E_x  = qx / rho * (E + p)
  flx_E_y  = qy / rho * (E + p)
  @returns flx_rho_x, flx_rho_y, flx_qx_x, flx_qx_y, flx_qy_y, flx_E_x, flx_E_y
end


@with_signature function llf_rho(equation::EulerEquation2d)
  @accepts Prefix(lhs), rho, flx_rho_x, flx_rho_y, v
  @accepts Prefix(rhs), rho, flx_rho_x, flx_rho_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_rho_x + ny*lhs_flx_rho_y
  rhs_nflx = nx*rhs_flx_rho_x + ny*rhs_flx_rho_y
  vmax = max(lhs_v, rhs_v)
  nflx_rho = LLF(lhs_nflx, rhs_nflx, lhs_rho, rhs_rho, vmax)
  @returns nflx_rho
end
@with_signature function llf_qx(equation::EulerEquation2d)
  @accepts Prefix(lhs), qx, flx_qx_x, flx_qx_y, v
  @accepts Prefix(rhs), qx, flx_qx_x, flx_qx_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_qx_x + ny*lhs_flx_qx_y
  rhs_nflx = nx*rhs_flx_qx_x + ny*rhs_flx_qx_y
  vmax = max(lhs_v, rhs_v)
  nflx_qx = LLF(lhs_nflx, rhs_nflx, lhs_qx, rhs_qx, vmax)
  @returns nflx_qx
end
@with_signature function llf_qy(equation::EulerEquation2d)
  @accepts Prefix(lhs), qy, flx_qy_x, flx_qy_y, v
  @accepts Prefix(rhs), qy, flx_qy_x, flx_qy_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_qy_x + ny*lhs_flx_qy_y
  rhs_nflx = nx*rhs_flx_qy_x + ny*rhs_flx_qy_y
  vmax = max(lhs_v, rhs_v)
  nflx_qy = LLF(lhs_nflx, rhs_nflx, lhs_qy, rhs_qy, vmax)
  @returns nflx_qy
end
@with_signature function llf_E(equation::EulerEquation2d)
  @accepts Prefix(lhs), E, flx_E_x, flx_E_y, v
  @accepts Prefix(rhs), E, flx_E_x, flx_E_y, v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_E_x + ny*lhs_flx_E_y
  rhs_nflx = nx*rhs_flx_E_x + ny*rhs_flx_E_y
  vmax = max(lhs_v, rhs_v)
  nflx_E = LLF(lhs_nflx, rhs_nflx, lhs_E, rhs_E, vmax)
  @returns nflx_E
end


@with_signature function maxspeed(equation::EulerEquation2d)
  @accepts rho, qx, qy, E
  @unpack eos = equation
  q = sqrt(qx^2+qy^2)
  eps, = internal_energy((rho,q,E), equation)
  p = eos(Pressure, Density, rho, InternalEnergy, eps)
  v = abs(q / p) + sqrt(abs(eos.gamma * p / rho))
  @returns v
end


#######################################################################
#                         LDG implementation                          #
#######################################################################


@with_signature function ldg_flux(equation::EulerEquation)
  @accepts State(ldg_rho, ldg_q, ldg_E), rho, q, E
  flx_ldg_rho, flx_ldg_q, flx_ldg_E = rho, q, E
  @returns flx_ldg_rho, flx_ldg_q, flx_ldg_E
end


@with_signature function ldg_speed(equation::EulerEquation)
  @accepts State(ldg_rho, ldg_q, ldg_E), rho, q, E
  v_ldg_rho, v_ldg_q, v_ldg_E = 1, 1, 1
  @returns v_ldg_rho, v_ldg_q, v_ldg_E
end


@with_signature function av_flux(equation::EulerEquation)
  @accepts State(ldg_rho, ldg_q, ldg_E), smoothed_mu
  flx_ldg_rho, flx_ldg_q, flx_ldg_E =
    smoothed_mu * ldg_rho, smoothed_mu * ldg_q, smoothed_mu * ldg_E
  @returns flx_ldg_rho, flx_ldg_q, flx_ldg_E
end


@with_signature function av_speed(equation::EulerEquation)
  @accepts State(ldg_rho, ldg_q, ldg_E), smoothed_mu
  v_av_rho, v_av_q, v_av_E = smoothed_mu, smoothed_mu, smoothed_mu
  @returns v_av_rho, v_av_q, v_av_E
end
