initialdata!(env, P, prms) = initialdata!(env, P, prms, env.mesh)

function initialdata!(env, P::Project, prms, mesh::Mesh1d)
  initialdata_equation_of_state!(env, P, P.equation.eos, prms, mesh)
  initialdata_entropy_variables!(env, P, prms)
  initialdata_hrsc!(env, P, P.hrsc, prms)
  initialdata_tci!(env, P, P.tci, prms)
end

function initialdata!(env, P::Project2d, prms, mesh::Mesh2d)
  initialdata_equation_of_state!(env, P, P.equation.eos, prms, mesh)
end


function initialdata_equation_of_state!(env, P::Project, eos::EquationOfState.IdealGas, prms, ::Mesh1d)

  @unpack id = prms

  @unpack cache, mesh = env

  (xmin, xmax), = extrema(mesh)
  γ = eos.gamma
  x = vec(mesh.x)

  E(p, rho, v) = p / (γ - 1.0) + 0.5 * rho * v^2

  tuple_data = if id == "smooth_isentropic_flow"
    # Glaubitz, PhD thesis 2020, p. 219
    @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
    @toggled_assert γ == 3.0
    ρ0 = @. 1.0 + 0.5 * sin(π * x)
    v0 = @. 0.0 * x
    p0 = @. ρ0^γ
    q0 = @. ρ0 * v0
    E0 = @. E(p0, ρ0, v0)
    (ρ0, q0, E0, p0)
  elseif id == "sod_shock_tube"
    @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
    @toggled_assert γ == 1.4
    ρ0 = [ (xi < 0.0) ? 1.0 : 0.125  for xi in x ]
    v0 = [ (xi < 0.0) ? 0.0 : 0.0    for xi in x ]
    p0 = [ (xi < 0.0) ? 1.0 : 0.1    for xi in x ]
    q0 = @. ρ0 * v0
    E0 = @. E(p0, ρ0, v0)
    (ρ0, q0, E0, p0)
  elseif id == "shu_osher_shock_tube_v1"
    # taken from Glaubitz 2020, PhD thesis
    # results do not agree, I believe that ID is wrong
    @toggled_assert isapprox(xmin, 0.0) && isapprox(xmax, 4.0)
    @toggled_assert γ == 1.4
    ρ0 = [ (xi <= 2.5) ? 3.857143 : 1.0 + 0.2 * sin(10*π*xi) for xi in x ]
    v0 = [ (xi <= 2.5) ? 2.629369 : 0.0    for xi in x ]
    p0 = [ (xi <= 2.5) ? 10.333333 : 1.0   for xi in x ]
    q0 = @. ρ0 * v0
    E0 = @. E(p0, ρ0, v0)
    (ρ0, q0, E0, p0)
  elseif id == "shu_osher_shock_tube_v2"
    # taken from Yu, Hesthaven, 2019
    @toggled_assert isapprox(xmin, -5.0) && isapprox(xmax, 5.0)
    @toggled_assert γ == 1.4
    ρ0 = [ (xi <= -4.0) ? 3.857143 : 1.0 + 0.2 * sin(5*xi) for xi in x ]
    v0 = [ (xi <= -4.0) ? 2.629369 : 0.0    for xi in x ]
    p0 = [ (xi <= -4.0) ? 10.333333 : 1.0    for xi in x ]
    q0 = @. ρ0 * v0
    E0 = @. E(p0, ρ0, v0)
    (ρ0, q0, E0, p0)
  else
    error("Initial data of type $id not implemented!")
  end

  # pack data
  @unpack rho, q, E = get_dynamic_variables(cache)
  @unpack p         = get_static_variables(cache)
  rho .= tuple_data[1]
  q   .= tuple_data[2]
  E   .= tuple_data[3]
  p   .= tuple_data[4]

  return
end


function initialdata_equation_of_state!(env, P, eos::EquationOfState.IdealGas, prms, ::Mesh2d)

  @unpack id = prms
  @unpack cache, mesh = env
  @unpack x, y = get_static_variables(cache)

  (xmin, xmax), (ymin, ymax) = extrema(mesh)
  γ = eos.gamma

  E(p, rho, v) = p / (γ - 1.0) + 0.5 * rho * v^2
  tuple_data = if id == "smooth_isentropic_flow"
    # Glaubitz, PhD thesis 2020, p. 219
    @toggled_assert isapprox(xmin, -1.0) && isapprox(xmax, 1.0)
    @toggled_assert γ == 3.0
    ρ0 = @. 1.0 + 0.5 * sin(π * x)
    vx0 = @. 0.0 * x
    vy0 = @. 0.0 * x
    p0 = @. ρ0^γ
    qx0 = @. ρ0 * vx0
    qy0 = @. ρ0 * vy0
    v0 = @. sqrt(vx0^2 + vy0^2)
    E0 = @. E(p0, ρ0, v0)
    (ρ0, qx0, qy0, E0, p0)
  else
    error("Initial data of type $id not implemented!")
  end

  # pack data
  @unpack rho, qx, qy, E = get_dynamic_variables(cache)
  @unpack p              = get_static_variables(cache)
  rho .= tuple_data[1]
  qx  .= tuple_data[2]
  qy  .= tuple_data[3]
  E   .= tuple_data[4]
  p   .= tuple_data[5]

  return
end


function initialdata_entropy_variables!(env, P::Project, prms)
  @unpack cache = env
  @unpack equation = P
  @unpack t, tm1 = get_global_variables(cache)
  @unpack S, Sm1, flx_S, flx_Sm1 = get_static_variables(cache)
  tm1 .= 0.0
  t   .= 0.0
  broadcast_volume!(entropy, equation, cache)
  broadcast_volume!(entropy_flux, equation, cache)
  @. Sm1 = S
  @. flx_Sm1 = flx_S
end


function initialdata_hrsc!(env, P::Project, hrsc::Maybe{HRSC.AbstractHRSC}, prms) end


function initialdata_hrsc!(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, prms)
  @unpack mu = get_cell_variables(env.cache)
  @. mu = 0.0
end


function initialdata_tci!(env, P::Project, tci::Nothing, prms) end


function initialdata_tci!(env, P::Project, tci::TCI.EntropyProduction, prms)
  @unpack cache = env
  @unpack EP = get_static_variables(cache)
  @unpack EPjump_lhs, EPjump_rhs = get_bdry_variables(cache)
  @. EP = 0.0
  @. EPjump_lhs = 0.0
  @. EPjump_rhs = 0.0
end


function initialdata_tci!(env, P::Project, tci::TCI.AbstractTCI, prms)
  @unpack flag = get_cell_variables(env.cache)
  @. flag = 0.0
end
