#######################################################################
#                              Timestep                               #
#######################################################################


function timestep(env, P::Project, hrsc::Maybe{HRSC.AbstractHRSC}, ::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume!(speed, equation, cache)
  v = get_variable(cache, :v)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL * dl / (N * vmax)
  return dt
end


function timestep(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, ::Mesh1d)
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume!(speed, equation, cache)
  v = get_variable(cache, :v)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end
  smoothed_mu = get_variable(cache, :smoothed_mu)
  mumax = dg1d.absolute_maximum(view(smoothed_mu, :))

  @unpack dl, x, CFL, element = mesh
  @unpack N = element

  dt = CFL / (vmax * N / dl + mumax * N^3 / dl^2)
  dt_limit = 1e-7
  if dt < dt_limit
    @warn "Limiting timestep due to dt being smaller than $dt_limit"
    dt = dt_limit
  end
  return dt
end


function timestep(env, P::Project2d, hrsc::Maybe{HRSC.AbstractHRSC}, ::Mesh2d)
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume_2!(maxspeed, equation, mesh)
  v = get_variable(cache, :v)
  vmax = dg1d.absolute_maximum(view(v,:))
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack CFL, element = mesh
  @unpack N, z = element

  dx, dy = dg1d.widths(mesh.boxes[1])
  for box in view(mesh.boxes, 2:length(mesh.boxes))
    ddx, ddy = dg1d.widths(box)
    ddx < dx && (dx = ddx)
    ddy < dy && (dy = ddy)
  end
  dz = z[2]-z[1]
  dl = min(dx, dy) * dz

  dt = CFL * dl / (N * vmax)

  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


function rhs!(env, P::Project, hrsc::Nothing, ::Mesh1d, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh                   = env
  @unpack equation, rsolver             = P
  @unpack rho, q, E                     = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E         = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q , nflx_E     = get_bdry_variables(cache)

  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  # broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   nflx_E,   mesh)

  return
end


function rhs!(env, P::Project2d, hrsc::Nothing, ::Mesh2d, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh                        = env
  @unpack equation, rsolver_rho, rsolver_qx,
          rsolver_qy, rsolver_E              = P
  @unpack rho, qx, qy, E                     = get_dynamic_variables(cache)
  @unpack flx_rho_x, flx_rho_y,
          flx_qx_x, flx_qx_y,
          flx_qy_x, flx_qy_y,
          flx_E_x, flx_E_y, p, eps           = get_static_variables(cache)
  @unpack rhs_rho, rhs_qx, rhs_qy, rhs_E     = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_qx, nflx_qy, nflx_E = get_bdry_variables(cache)

  broadcast_volume_2!(flux, equation, mesh)
  broadcast_faces_2!(llf_rho, equation, mesh)
  broadcast_faces_2!(llf_qx,  equation, mesh)
  broadcast_faces_2!(llf_qy,  equation, mesh)
  broadcast_faces_2!(llf_E,   equation, mesh)

  compute_rhs_weak_form!(rhs_rho, flx_rho_x, flx_rho_y,  nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_qx,  flx_qx_x,  flx_qx_y,   nflx_qx,  mesh)
  compute_rhs_weak_form!(rhs_qy,  flx_qy_x,  flx_qy_y,   nflx_qy,  mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E_x,   flx_E_y,    nflx_E,   mesh)

  return
end


function rhs!(env, P::Project, hrsc::AbstractReconstruction, ::Mesh1d, bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh                   = env
  @unpack equation, rsolver             = P
  @unpack rho, q, E                     = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps = get_static_variables(cache)
  @unpack flag                          = get_cell_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E         = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q, nflx_E      = get_bdry_variables(cache)

  HRSC.reconstruct!(rho,  flag, hrsc, isperiodic=isperiodic(bdryconds))
  HRSC.reconstruct!(q,    flag, hrsc, isperiodic=isperiodic(bdryconds))
  HRSC.reconstruct!(E,    flag, hrsc, isperiodic=isperiodic(bdryconds))

  broadcast_volume!(flux, equation, cache)
  broadcast_volume!(speed, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  # broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   nflx_E,   mesh)

  return
end


function rhs!(env, P::Project, hrsc::HRSC.AbstractArtificialViscosity, ::Mesh1d,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh = env
  @unpack equation, rsolver, ldg_rsolver, av_rsolver = P

  @unpack rho, q, E                            = get_dynamic_variables(cache)
  @unpack flx_rho, flx_q, flx_E, p, eps,
          ldg_rho, ldg_q, ldg_E,
          flx_ldg_rho, flx_ldg_q, flx_ldg_E    = get_static_variables(cache)
  @unpack rhs_rho, rhs_q, rhs_E                = get_rhs_variables(cache)
  @unpack nflx_rho, nflx_q, nflx_E,
          nflx_ldg_rho, nflx_ldg_q, nflx_ldg_E = get_bdry_variables(cache)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_volume!(ldg_flux, equation, cache)
  broadcast_faces!(central_flux, ldg_rsolver, cache, mesh)
  # broadcast_boundaryconditions!(central_flux, ldg_bdryconds, cache, mesh, 0.0)
  compute_rhs_weak_form!(ldg_rho, flx_ldg_rho,  nflx_ldg_rho, nflx_ldg_rho, mesh)
  compute_rhs_weak_form!(ldg_q,   flx_ldg_q,    nflx_ldg_q,   nflx_ldg_q,   mesh)
  compute_rhs_weak_form!(ldg_E,   flx_ldg_E,    nflx_ldg_E,   nflx_ldg_E,   mesh)

  ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(av_flux, equation, cache)
  broadcast_faces!(mean_flux, av_rsolver, cache, mesh)
  # broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, 0.0)
  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  # broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  @. flx_rho  += flx_ldg_rho
  @. flx_q    += flx_ldg_q
  @. flx_E    += flx_ldg_E
  @. nflx_rho += nflx_ldg_rho
  @. nflx_q   += nflx_ldg_q
  @. nflx_E   += nflx_ldg_E

  compute_rhs_weak_form!(rhs_rho, flx_rho,  nflx_rho, nflx_rho, mesh)
  compute_rhs_weak_form!(rhs_q,   flx_q,    nflx_q,   nflx_q,   mesh)
  compute_rhs_weak_form!(rhs_E,   flx_E,    nflx_E,   nflx_E,   mesh)

  return
end
