Project(env::Environment, prms) = Project(env, env.mesh, prms)


function Project(env::Environment, mesh::Mesh1d, prms)

  # construct Project
  eos = EquationOfState.make_EquationOfState(prms["EquationOfState"])
  equation = EulerEquation(eos)
  tci  = TCI.make_TCI(env.mesh, prms["TCI"])
  hrsc = HRSC.make_HRSC(env.mesh, prms["HRSC"])
  rsolver = ApproxRiemannSolver(flux, speed, equation)
  ldg_rsolver, av_rsolver = if hrsc isa AbstractArtificialViscosity
    ApproxRiemannSolver(ldg_flux, ldg_speed, equation),
    ApproxRiemannSolver(av_flux, av_flux, equation)
  else
    nothing, nothing
  end

  P = Project(equation, rsolver, hrsc, tci, ldg_rsolver, av_rsolver)

  # register variables
  # TODO add a ::Nothing overload for register_variables!
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(cache, P)
  register_variables!(cache, rsolver)
  !isnothing(ldg_rsolver) && register_variables!(cache, ldg_rsolver)
  !isnothing(av_rsolver) && register_variables!(cache, av_rsolver, dont_register=true)
  display(cache)

  # setup initial data
  initialdata!(env, P, prms["EulerEq"])

  # setup boundary conditions
  bdryconds = make_BoundaryConditions(env, equation, rsolver, prms["EulerEq"])
  ldg_bdryconds, av_bdryconds = if hrsc isa AbstractArtificialViscosity
    make_BoundaryConditions_LDG(env, equation, ldg_rsolver, prms["EulerEq"]),
    make_BoundaryConditions_LDG(env, equation, av_rsolver, prms["EulerEq"])
  else
    nothing, nothing
  end

  # TODO Bdry conditions should be moved to project and should only be just a 'lightweight'
  # descriptor. Furthermore, they should also register variables in cache themselves.
  # Note: Atm, only the 'from_initial_data' option would require a place to store the data
  # somewhere. For periodic BCs we don't need extra storage at all.
  # But what other types of boundary conditions will be needed in the future?

  # setup callbacks
  projectcb = make_callback(env, P, bdryconds)

  append!(env.callbacks, CallbackSet(projectcb.callbacks...))

  return P, (bdryconds, ldg_bdryconds, av_bdryconds)
end


function Project(env::Environment, mesh::Mesh2d, prms)

  # construct Project
  eos = EquationOfState.make_EquationOfState(prms["EquationOfState"])
  equation = EulerEquation2d(eos)
  ldg_rsolver, av_rsolver = nothing, nothing

  P = Project2d(equation, nothing, nothing,
                nothing, nothing, nothing, nothing, nothing, nothing)

  # register variables
  # TODO Somehow replace _register_variables! with register_variables!
  @unpack cache = env
  _register_variables!(cache, P, mesh)
  display(cache)

  # setup initial data
  initialdata!(env, P, prms["EulerEq"])

  # setup boundary conditions
  bdryconds_x = nothing
  bdryconds_y = nothing
  ldg_bdryconds, av_bdryconds = nothing, nothing

  return P, ((bdryconds_x, bdryconds_y), ldg_bdryconds, av_bdryconds)
end


#######################################################################
#                         Register variables                          #
#######################################################################


# Julia limitation/bug: Overloading imported functions without their Module prefix shadows their
# definition in this module without a warning or error. This does not happen for methods
# defined in Base.
# Because of that we prefix with _ below.
# Cf. https://github.com/JuliaLang/julia/issues/15510
function _register_variables!(cache, P::Project)
  register_variables!(cache,
    dynamic_variablenames = (:rho, :q, :E,),     # density, momentum density, energy
    rhs_variablenames     = (:rhs_rho, :rhs_q, :rhs_E),
    static_variablenames  = (:flx_rho, :flx_q, :flx_E,
                             :p, :eps, :v,        # pressure, internal energy, speed
                             :S, :Sm1,           # entropy
                             :flx_S, :flx_Sm1),  # entropy flux
    cell_variablenames    = (:max_v,),
    global_variablenames  = (:t, :tm1) # recent time stamps
  )
  _register_variables!(cache, P.hrsc)
  _register_variables!(cache, P.tci)
end


_register_variables!(cache, tci_or_hrsc::Nothing) = nothing

_register_variables!(cache, hrsc::AbstractReconstruction) = nothing


function _register_variables!(cache, hrsc::HRSC.AbstractArtificialViscosity)
  register_variables!(P.cache,
    static_variablenames  = (:ldg_rho, :ldg_q, :ldg_E,              # local DG variable
                             :flx_ldg_rho, :flx_ldg_q, :flx_ldg_E), # local DG flux
    cell_variablenames    = (:mu,) # 'one viscosity to rule them all'
    )
end


function _register_variables!(cache, hrsc::HRSC.SmoothedArtificialViscosity)
  _register_variables!(cache, hrsc.av)
  register_variables!(cache,
    static_variablenames  = (:smoothed_mu,)
  )
end


function _register_variables!(cache, tci::TCI.AbstractTCI)
  register_variables!(cache, cell_variablenames = (:flag,))
end


function _register_variables!(cache, tci::TCI.EntropyProduction)
  register_variables!(cache,
                      cell_variablenames = (:flag,),
                      static_variablenames = (:EP,), # entropy production
                      bdry_variablenames   = (:EPjump_lhs, :EPjump_rhs)
                                              # entropy production due to jumps at interfaces
  )
end


function _register_variables!(cache, P::Project2d, ::Mesh2d)
  register_variables!(cache,
    dynamic_variablenames = (:rho, :qx, :qy, :E),     # density, momentum density, energy
    rhs_variablenames     = (:rhs_rho, :rhs_qx, :rhs_qy, :rhs_E),
    static_variablenames  = (:flx_rho_x, :flx_rho_y,
                             :flx_qx_x, :flx_qx_y,
                             :flx_qy_x, :flx_qy_y,
                             :flx_E_x, :flx_E_y,
                             :p, :eps, :v),        # pressure, internal energy, speed
    bdry_variablenames    = (:nflx_rho, :nflx_qx, :nflx_qy, :nflx_E),
    cell_variablenames    = (:max_v,),
    global_variablenames  = (:t,) # recent time stamps
  )
end
