function make_Equation(eos, prms, m::Mesh)
  @unpack atm_density, atm_threshold_factor = prms
  atm = Atmosphere(atm_density, atm_threshold_factor)
  cons2prim_newtonsolver = dg1d.NewtonSolver(approx_derivative=true)
  cons2prim_bisection = dg1d.Bisection()
  if m isa Mesh1d
    return Equation(eos, atm, cons2prim_newtonsolver, cons2prim_bisection)
  elseif m isa Mesh2d
    return Equation2d(eos, atm, cons2prim_newtonsolver, cons2prim_bisection)
  else
    error("unknown mesh type encountered")
  end
end


#######################################################################
#                           Implementation                            #
#######################################################################


module Formulae
  # auxilary variables: W, h, s
  Wprim(ρ, v, ϵ, p) = 1.0 / sqrt(1.0 - v^2) # in terms of primitive variables
  Wcons(D, S, τ, p) = (τ + p + D) / sqrt((τ + p + D)^2 - S^2) # in terms of conservative variables
  # inverse relations, see Bugner et al. (2018), see my notes
  ρ(D, S, τ, p) = D / Wcons(D, S, τ, p)
  v(D, S, τ, p) = S / (τ + D + p)
  ϵ(D, S, τ, p) = (sqrt((τ + p + D)^2 - S^2) - Wcons(D, S, τ, p) * p) / D - 1
  # conservative variables D, S, τ in terms of primitive ones
  D(ρ, v, ϵ, p) = ρ * Wprim(ρ, v, ϵ, p)
  S(ρ, v, ϵ, p) = ρ * (1 + ϵ + p / ρ) * Wprim(ρ, v, ϵ, p)^2 * v
  τ(ρ, v, ϵ, p) = ρ * (1 + ϵ + p / ρ) * Wprim(ρ, v, ϵ, p)^2 - p - D(ρ, v, ϵ, p)
  # conservative fluxes
  flux_D(ρ, v, ϵ, p) = D(ρ, v, ϵ, p) * v
  flux_S(ρ, v, ϵ, p) = S(ρ, v, ϵ, p) * v + p
  flux_τ(ρ, v, ϵ, p) = (τ(ρ, v, ϵ, p) + p) * v
end


@with_signature function flux(equation::Equation)
  @accepts State(D, S, tau), p, rho, v, eps
  flx_D   = Formulae.flux_D(rho, v, eps, p)
  flx_S   = Formulae.flux_S(rho, v, eps, p)
  flx_tau = Formulae.flux_τ(rho, v, eps, p)
  @returns flx_D, flx_S, flx_tau
end


@with_signature function speed(equation::Equation)
  @accepts State(D, S, tau), p, rho, v, eps
  @unpack eos = equation
  cs2 = eos(SquareSpeedOfSound, Density, rho, InternalEnergy, eps)
  cs = sqrt(abs(cs2)) # abs because cs2 might be negative due to cons2prim_insane
  v0 = v
  vp = (v + cs) / (1 + v * cs)
  vm = (v - cs) / (1 - v * cs)
  max_v = maximum(abs.((v0, vp, vm)))
  @returns max_v
end


# @with_signature function impose_atmosphere(equation::Equation)
#   @accepts rho, p, eps, v, D, S, tau
#   @unpack atmosphere = equation
#
#   ρatm = atmosphere.rho
#   patm = 0.0
#   ϵatm = tau / ρatm
#   hatm = 1.0
#   Watm = 1.0
#   vatm = 1.0
#
#   if rho < atmosphere.rho * atmosphere.threshold
#     rho = rhoatm
#     p   = patm
#     eps = epsatm
#     v   = vatm
#     D   = Formulae.D(rhoatm, vatm, epsatm, patm)
#     S   = Formulae.S(rhoatm, vatm, epsatm, patm)
#     tau = Formulae.τ(rhoatm, vatm, epsatm, patm)
#   end
#
#   @returns rho, p, eps, v, D, S, tau
# end


@with_signature function primitives(equation::Equation)
  @accepts D, S, tau, p
  rho = Formulae.ρ(D, S, tau, p)
  v   = Formulae.v(D, S, tau, p)
  eps = Formulae.ϵ(D, S, tau, p)
  @returns rho, v, eps
end


@with_signature function conservatives(equation::Equation)
  @accepts rho, v, eps, p
  D   = Formulae.D(rho, v, eps, p)
  S   = Formulae.S(rho, v, eps, p)
  tau = Formulae.τ(rho, v, eps, p)
  @returns D, S, tau
end


@with_signature function entropy_variables(equation::Equation{EquationOfState.IdealGas})
  @accepts rho, p, E, flx_E
  @unpack eos = equation
  γ     = eos.gamma
  E     = rho / (γ - 1) * log(p / rho^γ)
  flx_E = q * E / rho
  @returns E, flx_E
end


@with_signature function cons2prim(equation::Equation)
  @accepts D, S, tau, p, x
  rho, v, eps, p = cons2prim_insane((D, S, tau, p, x), equation)
  @returns rho, v, eps, p
end


@with_signature function cons2prim_insane(equation::Equation)
  @accepts D, S, tau, p, x
  (p,) = find_pressure_insane((D, S, tau, p, x), equation)
  rho  = Formulae.ρ(D, S, tau, p)
  v    = Formulae.v(D, S, tau, p)
  eps  = Formulae.ϵ(D, S, tau, p)
  @returns rho, v, eps, p
end


function pressure_equation_insane(tau, p, D, S, eos, p0, x)
  S2 = S^2
  a = tau + p + D
  a2mS2 = a^2 - S2
  if a2mS2 < 0
    return NaN
  end
  W = a / sqrt(a2mS2)
  rho = D / W
  eps = (a - W^2 * p) / (W^2 * rho) - 1
  if eps < 0 # will break pressure computation below
    return NaN
  end
  eos_p = eos(Pressure, Density, rho, InternalEnergy, eps)
  f = eos_p - p
  return f
end


@with_signature function find_pressure_insane(equation::Equation)
  @accepts D, S, tau, p, x
  @unpack eos = equation

  # upper limit on particle velocity
  omv2 = 1e-3 # = 1 - v^2

  S2   = S^2
  # fails in relativistic shocktube 2
  # if tau <= -1.0 || D <= 0.0
  #   error("Invalid state encountered! (τ, D) = ($tau, $D)")
  # end
  pmin1 = - (tau + D) + sqrt(S2 + omv2) # √((τ + D + p)^2 - S^2) >= √(1 - vmax^2)
  pmin2 = 0 # physical constraint
  pmin = max(pmin1, pmin2)

  floor_pressure(pnew) = max(pnew, pmin)
  # p0 = 0.5 * (eos.gamma - 1) * (tau + D)
  p0 = p

  f = pp -> pressure_equation_insane(tau, pp, D, S, eos, p0, x)
  @unpack cons2prim_newtonsolver, cons2prim_bisection = equation
  p, err, converged, tol = find_root(f, p0, cons2prim_newtonsolver, sanitize=floor_pressure)
  if !converged
    p = pmin
  end

  @returns p
end


#######################################################################
#                                 2d                                  #
#######################################################################


module Formulae2d
  # auxilary variables: W, h, s
  Wprim(ρ, vx, vy, ϵ, p) = 1.0 / sqrt(1.0 - (vx^2+vy^2)) # in terms of primitive variables
  Wcons(D, Sx, Sy, τ, p) = (τ + p + D) / sqrt((τ + p + D)^2 - (Sx^2+Sy^2)) # in terms of conservative variables
  # inverse relations, see Bugner et al. (2018), see my notes
  ρ(D, Sx, Sy, τ, p)  = D / Wcons(D, Sx, Sy, τ, p)
  vx(D, Sx, Sy, τ, p) = Sx / (τ + D + p)
  vy(D, Sx, Sy, τ, p) = Sy / (τ + D + p)
  ϵ(D, Sx, Sy, τ, p)  = (sqrt((τ + p + D)^2 - (Sx^2 + Sy^2)) - Wcons(D, Sx, Sy, τ, p) * p) / D - 1
  # conservative variables D, S, τ in terms of primitive ones
  D(ρ, vx, vy, ϵ, p) = ρ * Wprim(ρ, vx, vy, ϵ, p)
  Sx(ρ, vx, vy, ϵ, p) = ρ * (1 + ϵ + p / ρ) * Wprim(ρ, vx, vy, ϵ, p)^2 * vx
  Sy(ρ, vx, vy, ϵ, p) = ρ * (1 + ϵ + p / ρ) * Wprim(ρ, vx, vy, ϵ, p)^2 * vy
  τ(ρ, vx, vy, ϵ, p) = ρ * (1 + ϵ + p / ρ) * Wprim(ρ, vx, vy, ϵ, p)^2 - p - D(ρ, vx, vy, ϵ, p)
end


@with_signature function cons2prim(equation::Equation2d)
  @accepts D, Sx, Sy, tau, x, y, p
  S = sqrt(Sx^2 + Sy^2)
  p, = find_pressure_insane((D, S, tau, p, x, y), equation)
  rho = Formulae2d.ρ(D,  Sx, Sy, tau, p)
  vx  = Formulae2d.vx(D, Sx, Sy, tau, p)
  vy  = Formulae2d.vy(D, Sx, Sy, tau, p)
  eps = Formulae2d.ϵ(D,  Sx, Sy, tau, p)
  @returns rho, vx, vy, eps, p
end


@with_signature function find_pressure_insane(equation::Equation2d)
  @accepts D, S, tau, p, x, y
  @unpack eos = equation

  # upper limit on particle velocity
  omv2 = 1e-3 # = 1 - v^2

  S2   = S^2
  # fails in relativistic shocktube 2
  # if tau <= -1.0 || D <= 0.0
  #   error("Invalid state encountered! (τ, D) = ($tau, $D)")
  # end
  pmin1 = - (tau + D) + sqrt(S2 + omv2) # √((τ + D + p)^2 - S^2) >= √(1 - vmax^2)
  pmin2 = 0 # physical constraint
  pmin = max(pmin1, pmin2)

  floor_pressure(pnew) = max(pnew, pmin)
  # p0 = 0.5 * (eos.gamma - 1) * (tau + D)
  p0 = p

  f = pp -> pressure_equation_insane(tau, pp, D, S, eos, p0, (x,y))
  @unpack cons2prim_newtonsolver, cons2prim_bisection = equation
  p, err, converged, tol = find_root(f, p0, cons2prim_newtonsolver, sanitize=floor_pressure)
  if !converged
    p = pmin
  end

  @returns p
end


@with_signature function flux(equation::Equation2d)
  @accepts D, Sx, Sy, rho, eps, vx, vy, p, tau
  h = 1.0 + eps + p / rho
  v2 = vx^2 + vy^2
  ooW2 = 1.0 - v2
  flx_D_x = D * vx
  flx_D_y = D * vy
  flx_Sx_x = rho * h / ooW2 * vx*vx + p
  flx_Sx_y = rho * h / ooW2 * vx*vy
  flx_Sy_x = flx_Sx_y
  flx_Sy_y = rho * h / ooW2 * vy*vy + p
  flx_tau_x = (tau + p) * vx
  flx_tau_y = (tau + p) * vy
  @returns flx_D_x, flx_D_y, flx_Sx_x, flx_Sx_y, flx_Sy_y, flx_Sy_x, flx_Sy_y,
           flx_tau_x, flx_tau_y
end


@with_signature function maxspeed(equation::Equation2d)
  @accepts rho, vx, vy, eps
  @unpack eos = equation
  cs2 = eos(SquareSpeedOfSound, Density, rho, InternalEnergy, eps)
  cs = sqrt(abs(cs2)) # abs because cs2 might be negative due to cons2prim_insane
  v  = sqrt(vx^2 + vy^2)
  v0 = v
  vp = (v + cs) / (1 + v * cs)
  vm = (v - cs) / (1 - v * cs)
  max_v = maximum(abs.((v0, vp, vm)))
  @returns max_v
end


@inline LLF(nfl,nfr,ul,ur,v) = 1/2 * (nfl + nfr - v * (ur - ul))


@with_signature function llf_D(equation::Equation2d)
  @accepts Prefix(lhs), D, flx_D_x, flx_D_y, max_v
  @accepts Prefix(rhs), D, flx_D_x, flx_D_y, max_v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_D_x + ny*lhs_flx_D_y
  rhs_nflx = nx*rhs_flx_D_x + ny*rhs_flx_D_y
  vmax = max(lhs_max_v, rhs_max_v)
  nflx_D = LLF(lhs_nflx, rhs_nflx, lhs_D, rhs_D, vmax)
  @returns nflx_D
end
@with_signature function llf_Sx(equation::Equation2d)
  @accepts Prefix(lhs), Sx, flx_Sx_x, flx_Sx_y, max_v
  @accepts Prefix(rhs), Sx, flx_Sx_x, flx_Sx_y, max_v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_Sx_x + ny*lhs_flx_Sx_y
  rhs_nflx = nx*rhs_flx_Sx_x + ny*rhs_flx_Sx_y
  vmax = max(lhs_max_v, rhs_max_v)
  nflx_Sx = LLF(lhs_nflx, rhs_nflx, lhs_Sx, rhs_Sx, vmax)
  @returns nflx_Sx
end
@with_signature function llf_Sy(equation::Equation2d)
  @accepts Prefix(lhs), Sy, flx_Sy_x, flx_Sy_y, max_v
  @accepts Prefix(rhs), Sy, flx_Sy_x, flx_Sy_y, max_v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_Sy_x + ny*lhs_flx_Sy_y
  rhs_nflx = nx*rhs_flx_Sy_x + ny*rhs_flx_Sy_y
  vmax = max(lhs_max_v, rhs_max_v)
  nflx_Sy = LLF(lhs_nflx, rhs_nflx, lhs_Sy, rhs_Sy, vmax)
  @returns nflx_Sy
end
@with_signature function llf_tau(equation::Equation2d)
  @accepts Prefix(lhs), tau, flx_tau_x, flx_tau_y, max_v
  @accepts Prefix(rhs), tau, flx_tau_x, flx_tau_y, max_v
  @accepts nx, ny
  lhs_nflx = nx*lhs_flx_tau_x + ny*lhs_flx_tau_y
  rhs_nflx = nx*rhs_flx_tau_x + ny*rhs_flx_tau_y
  vmax = max(lhs_max_v, rhs_max_v)
  nflx_tau = LLF(lhs_nflx, rhs_nflx, lhs_tau, rhs_tau, vmax)
  @returns nflx_tau
end


#######################################################################
#                                 LDG                                 #
#######################################################################


@with_signature function ldg_flux(equations::Equation)
  @accepts State(D, S, tau)
  flx_ldg_D, flx_ldg_S, flx_ldg_tau = D, S, tau
  @returns flx_ldg_D, flx_ldg_S, flx_ldg_tau
end


@with_signature function ldg_speed(equations::Equation)
  @accepts State(D, S, tau)
  v_ldg_D, v_ldg_S, v_ldg_tau = 1, 1, 1
  @returns v_ldg_D, v_ldg_S, v_ldg_tau
end


@with_signature function av_flux(equations::Equation)
  @accepts State(ldg_D, ldg_S, ldg_tau), smoothed_mu
  flx_ldg_D, flx_ldg_S, flx_ldg_tau =
    smoothed_mu * ldg_D, smoothed_mu * ldg_S, smoothed_mu * ldg_tau
  @returns flx_ldg_D, flx_ldg_S, flx_ldg_tau
end


@with_signature function av_speed(equations::Equation)
  @accepts State(ldg_D, ldg_S, ldg_tau), smoothed_mu
  flx_ldg_D, flx_ldg_S, flx_ldg_tau = smoothed_mu, smoothed_mu, smoothed_mu
  @returns v_flx_ldg_D, v_flx_ldg_S, v_flx_ldg_tau
end
