#######################################################################
#                               Factory                               #
#######################################################################


function make_Equation(prms, dims)
  @unpack equation = prms
  if dims == 1
    if equation == "advection"
      v = prms["advection_v"]
      @toggled_assert length(v) == dims
      return Advection(v isa Vector ? v[1] : v)
    elseif equation == "burgers"
      return Burgers()
    else
    end
  elseif dims == 2
    if equation == "advection"
      v = prms["advection_v"]
      @toggled_assert length(v) == dims
      return Advection2d(v...)
    else
      return Burgers2d()
    end
  end
  error("Unknown equation requested: '$equation'")
end


#######################################################################
#                              Advection                              #
#######################################################################


@with_signature function flux(eq::Advection)
  @accepts State(u)
  flx_u = eq.v * u
  @returns flx_u
end

@with_signature function speed(eq::Advection)
  @accepts State(u)
  v = eq.v
  @returns v
end

@with_signature function entropy(eq::Advection)
  @accepts u
  E = u^2 / 2
  @returns E
end

@with_signature function entropy_flux(eq::Advection)
  @accepts u
  F = u^2 / 2
  @returns F
end


#######################################################################
#                            Advection 2d                             #
#######################################################################


@with_signature function flux(eq::Advection2d)
  @accepts State(u)
  flx_x, flx_y = eq.v_x * u, eq.v_y * u
  @returns flx_x, flx_y
end

@with_signature function speed(eq::Advection2d)
  @accepts State(u)
  v = sqrt(eq.v_x^2 + eq.v_y^2)
  @returns v
end


#######################################################################
#                               Burgers                               #
#######################################################################


entropy!(E, u, eq::Burgers, mesh::Mesh) = @. E = u^2 / 2

entropyflux!(F, u, eq::Burgers, mesh::Mesh) = @. F = u^3 * 2 / 3

@with_signature function flux(eq::Burgers)
  @accepts State(u)
  flx_u = u^2 / 2
  @returns flx_u
end

@with_signature function speed(eq::Burgers)
  @accepts State(u)
  v = u
  @returns v
end

@with_signature function entropy(eq::Burgers)
  @accepts u
  E = u^2 / 2
  @returns E
end

@with_signature function entropy_flux(eq::Burgers)
  @accepts u
  F = u^3 * 2 / 3
  @returns F
end


#######################################################################
#                              Burgers2d                              #
#######################################################################


@with_signature function flux(eq::Burgers2d)
  @accepts State(u)
  flx_x, flx_y = u^2/2, u^2/2
  @returns flx_x, flx_y
end

@with_signature function speed(eq::Burgers2d)
  @accepts State(u)
  v = sqrt(2) * u
  @returns v
end


#######################################################################
#                                 LDG                                 #
#######################################################################


@with_signature function ldg_flux(eq::AbstractScalarEq)
  @accepts State(q), u
  flx_q = u
  @returns flx_q
end

@with_signature function ldg_speed(eq::AbstractScalarEq)
  @accepts State(q), u
  v_q = 1.0
  @returns v_q
end

@with_signature function av_flux(eq::AbstractScalarEq)
  @accepts State(q), smoothed_mu
  flx_q = smoothed_mu * q
  @returns flx_q
end

@with_signature function av_speed(eq::AbstractScalarEq)
  @accepts State(q), smoothed_mu
  v_av_q = smoothed_mu
  @returns v_av_q
end
