#######################################################################
#                              Timestep                               #
#######################################################################

function timestep(env, P::Project, ::Mesh1d, hrsc::Maybe{HRSC.AbstractHRSC})
  @unpack cache, mesh = env
  @unpack equation = P
  @unpack v = get_static_variables(cache)

  broadcast_volume!(speed, equation, cache)
  vmax = dg1d.absolute_maximum(v)
  vmax_limit = 1e4
  if vmax > vmax_limit
    @warn "Limiting timestep due to maximum speed exceeding $vmax_limit"
    vmax = vmax_limit
  end

  @unpack CFL, element = mesh
  @unpack N, z = element
  dz = z[2]-z[1]
  dx, = minimum(dg1d.widths(mesh.boxes[1]))
  dl = dx * dz

  dt = CFL * dl / (N * vmax)
  return dt
end


function timestep(env, P::Project, ::Mesh2d, hrsc::Maybe{HRSC.AbstractHRSC})
  @unpack cache, mesh = env
  @unpack equation = P

  broadcast_volume!(speed, equation, cache)

  @unpack v = get_static_variables(cache)
  vmax = dg1d.absolute_maximum(v)

  @unpack CFL, element = mesh
  @unpack N, z = element

  dx, dy = dg1d.widths(mesh.boxes[1])
  for box in view(mesh.boxes, 2:length(mesh.boxes))
    ddx, ddy = dg1d.widths(box)
    ddx < dx && (dx = ddx)
    ddy < dy && (dy = ddy)
  end
  dz = z[2]-z[1]
  dl = min(dx, dy) * dz

  dt = CFL * dl / (N * vmax)

  return dt
end


#######################################################################
#                           RHS computation                           #
#######################################################################


function rhs!(env, P::Project, hrsc, bdryconds...)
  return rhs!(env, env.mesh, P, hrsc, bdryconds...)
end


function rhs!(env, mesh::Mesh1d, P::Project, hrsc::Nothing,
    bdryconds, ldg_bdryconds::Nothing, av_bdryconds::Nothing)

  @unpack cache, mesh = env
  @unpack equation, rsolver = P

  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  # TODO Had to change from mean_flux to central flux, because
  # we moved the minus sign from the weak form into the normal vector
  # and that one is currently not applied when running with central flux
  # but only with mean flux.
  # I need to go through the definitions of those again and figure out if their
  # naming is even correct.
  broadcast_boundaryconditions!(mean_flux, bdryconds, cache, mesh, 0.0)

  @unpack flx_u  = get_static_variables(cache)
  @unpack u      = get_dynamic_variables(cache)
  @unpack rhs_u  = get_rhs_variables(cache)
  @unpack nflx_u = get_bdry_variables(cache)
  compute_rhs_weak_form!(rhs_u, flx_u, nflx_u, nflx_u, mesh)

  return
end


function rhs!(env, mesh::Mesh2d, P::Project, hrsc::Nothing,
    bdryconds, ldg_bdryconds::Nothing, av_bdryconds::Nothing)

  @unpack cache = env
  @unpack equation, rsolver = P

  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  # broadcast_boundaryconditions!(central_flux, bdryconds_x, cache, mesh, 0.0)
  # broadcast_boundaryconditions!(central_flux, bdryconds_y, cache, mesh, 0.0)

  @unpack flx_x, flx_y = get_static_variables(cache)
  @unpack u            = get_dynamic_variables(cache)
  @unpack rhs_u        = get_rhs_variables(cache)
  @unpack nflx_u = get_bdry_variables(cache)
  compute_rhs_weak_form!(rhs_u, flx_x, flx_y, nflx_u, mesh)

  return
end


function rhs!(env, mesh::Mesh1d, P::Project, hrsc::Union{BernsteinReconstruction, WENOReconstruction},
    bdryconds, ldg_bdryconds::Nothing, av_bdryconds::Nothing)

  @unpack cache, mesh       = env
  @unpack equation, rsolver = P
  @unpack flag                           = get_cell_variables(cache)
  @unpack flx_u                          = get_static_variables(cache)
  @unpack u                              = get_dynamic_variables(cache)
  @unpack rhs_u                          = get_rhs_variables(cache)
  @unpack lhs_numflx_u, rhs_numflx_u     = get_bdry_variables(cache)

  HRSC.reconstruct!(u, flag, hrsc, isperiodic=isperiodic(bdryconds))

  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  compute_rhs_weak_form!(rhs_u, flx_u, lhs_numflx_u, rhs_numflx_u, mesh)

  return
end


function rhs!(env, mesh::Mesh1d, P::Project, hrsc::HRSC.AbstractArtificialViscosity,
    bdryconds, ldg_bdryconds, av_bdryconds)

  @unpack cache, mesh = env
  @unpack equation, rsolver, ldg_rsolver, av_rsolver = P

  @unpack flx_u, flx_q, q = get_static_variables(cache)
  @unpack u               = get_dynamic_variables(cache)
  @unpack rhs_u           = get_rhs_variables(cache)
  @unpack nflx_u, nflx_q  = get_bdry_variables(cache)

  ## solve auxiliary equation: q + ∂x u = 0
  broadcast_volume!(ldg_flux, equation, cache)
  broadcast_faces!(mean_flux, ldg_rsolver, cache, mesh)
  # broadcast_boundaryconditions!(central_flux, ldg_bdryconds, cache, mesh, 0.0)
  compute_rhs_weak_form!(q, flx_q, nflx_q, nflx_q, mesh)

  # ## compute rhs of regularized equation: ∂t u + ∂x f + ∂x μ q = 0
  broadcast_volume!(av_flux, equation, cache)
  broadcast_faces!(mean_flux, av_rsolver, cache, mesh)
  # broadcast_boundaryconditions!(mean_flux, av_bdryconds, cache, mesh, 0.0)
  broadcast_volume!(flux, equation, cache)
  broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
  # broadcast_boundaryconditions!(lax_friedrich_flux, bdryconds, cache, mesh, 0.0)

  # add up fluxs and numerical fluxes
  @. flx_u  += flx_q
  @. nflx_u += nflx_q
  compute_rhs_weak_form!(rhs_u, flx_u, nflx_u, nflx_u, mesh)

  return
end
