#######################################################################
#                  Boundary Condition implementions                   #
#######################################################################


"""
    abstract type AbstractBC end

Supertype for boundary condition types. To be used with `BoundaryConditions` interface.
"""
abstract type AbstractBC end


"""
    DirichletBC(bdryval)

Dirichlet boundary condition, `bdryval` might be time dependent.
"""
struct DirichletBC <: AbstractBC 
  bdryval::Function
end
(bc::DirichletBC)(t) = bc.bdryval(t)


"""
Constructor for Dirichlet BC with constant boundary value.
"""
DirichletBC(bdry_vals::NTuple) = DirichletBC((t)->bdry_vals)


struct OutflowBC    <: AbstractBC end
struct InflowBC     <: AbstractBC end
struct PeriodicBC   <: AbstractBC end


#######################################################################
#                    BoundaryConditions interface                     #
#######################################################################


"""
    BoundaryConditions

Boundary condition interface for computing numerical fluxes at the domain boundaries.
Conditions are enforced weakly or strongly using an `AbstractRiemannSolver` and
its interface.

See: `ApproxRiemannSolver`.

---

    BoundaryConditions(lhs_bc<:AbstractBC, rhs_bc<:AbstractBC, rsolver<:AbstractRiemannSolver)

Assumes that `lhs_bc, lhs_bc` return the boundary state at the lhs/rhs domain ends.
See `AbstractBC` and its subtypes for available boundary condition types.

# Example

```julia
julia> # mesh, cache, rsolver already defined

julia> lhs_bc = DirichletBC(t -> 0.0)

julia> rhs_bc = OutflowBC()

julia> bdryconds = BoundaryConditions(lhs_bc, rhs_bc, rsolver)
```
"""
struct BoundaryConditions{T_BC_LHS<:AbstractBC,
                          T_BC_RHS<:AbstractBC,
                          T_RSolver<:AbstractRiemannSolver}
  lhs_bc::T_BC_LHS
  rhs_bc::T_BC_RHS
  rsolver::T_RSolver
  isperiodic::Bool
  function BoundaryConditions(lhs_bc, rhs_bc, rsolver)
    @assert !xor(lhs_bc isa PeriodicBC, rhs_bc isa PeriodicBC)
    isperiodic = lhs_bc isa PeriodicBC && rhs_bc isa PeriodicBC
    if !isperiodic
      n_accepts = length(accepts(rsolver.flux, rsolver.equation)[1])
      lhs_bc_iscomptabiel = if lhs_bc isa OutflowBC
        true
      else
        lhs_state = lhs_bc(0)
        @assert lhs_state isa Tuple "Boundary state must be a tuple"
        n_accepts == length(lhs_state)
      end
      rhs_bc_iscomptabiel = if rhs_bc isa OutflowBC
        true
      else
        rhs_state = rhs_bc(0)
        @assert rhs_state isa Tuple "Boundary state must be a tuple"
        n_accepts == length(rhs_state)
      end
      @assert lhs_bc_iscomptabiel && rhs_bc_iscomptabiel "Boundary conditions \
        must return the same number of arguments the flux function '$(rsolver.flux)' accepts"
    end
    return new{typeof(lhs_bc), typeof(rhs_bc), typeof(rsolver)}(lhs_bc, rhs_bc, 
                                                                rsolver, isperiodic)
  end
end


isperiodic(bdry::BoundaryConditions) = bdry.isperiodic
