const TO = TimerOutput()


"""
    main(parfile)

Run program specified by a `parfile` in TOML format.
"""
function main(parfile)

  # normalize parfile name
  parfile = abspath(parfile)

  project_name, prms = load_parameters(parfile)

  # TODO Use PrettyTables.jl for summary
  # TODO Dump versioninfo() to some file
  # TODO Dump Pkg.status() to some file
  # TODO Dump `git status` to some file
  # TODO Dump gethostname() and other system info
  # summary = summarize(prms)
  # println(summary)

  # backup any previous output dir and setup a fresh one
  # TODO Rename this to make_outputdir
  outputdir = mk_outputdir(parfile)
  cp(parfile, joinpath(outputdir, basename(parfile)))

  env, project, rhs_fn, timestep_fn = setup(project_name, prms, outputdir)

  # TODO summarize current setup once and log to screen
  # println(summarize(env.mesh))
  # println(summarize(env.cache))
  # println(summarize(env.callbacks))
  # println(summarize(env.callbacks_substep))
  # println(summarize(project))

  display(prms)
  # dump database summary to output folder
  open(joinpath(outputdir, "parameters.log"), "w") do f
    showparameters(f, prms, annotate=true)
  end
  # dump a parfile with current settings (including defaults)
  open(joinpath(outputdir, "all_parameters.toml"), "w") do f
    TOML.print(f, Dict(prms))
  end

  reset_timer!(TO)
  @timeit TO "Evolution" begin
    # Let's go!!!
    evolve!(env, rhs_fn, timestep_fn, prms)
  end

  # finish timer, progress bar
  timercb = env.callbacks.callbacks[end-1]
  timercb([], 0.0, 0, force=true)
  progresscb = env.callbacks.callbacks[end-1]
  progresscb([], 0.0, 0, force=true)

  return
end


function load_parameters(parfile)

  # read parameters
  if !isfile(parfile)
    error("Cannot locate parameter file '$parfile'")
  end

  user_prms = TOML.parsefile(parfile)
  # switch to a flat dict where keys of nested dicts are concatenated with dots
  user_prms = Dict(string(k) => flatten(v) for (k,v) in pairs(user_prms))

  ### sanitize parameter sections
  # supported top-level headers that aren't user projects
  default_projects = [ :Mesh, :Evolution, :Plot, :Output, :TCI, :HRSC, :EquationOfState ]

  for (k,v) in user_prms
    if !(v isa AbstractDict)
      error("Invalid parameter file: No top-level parameters allowed, found '$k = $v'")
    end
  end

  # find a parameter section for which somebody registered a project with that name
  all_projects = Symbol.(keys(user_prms))
  maybe_user_projects = filter(all_projects) do k
    return !(k in default_projects)
  end
  registered_projects = filter(isregistered ∘ Symbol, maybe_user_projects)
  not_registered_projects = setdiff(maybe_user_projects, registered_projects)
  if length(not_registered_projects) > 0
    error("Invalid parameter file: Found unknown top-level sections: $(not_registered_projects)")
  end
  if length(registered_projects) == 0
    error("Invalid parameter file: Missing a parameter section for a project, don't know which project to run ...")
  end
  if length(registered_projects) > 1
    error("Invalid parameter file: Only one project parameter section allowed, found multiple: $(registered_projects)")
  end

  # we found a user project
  project_name = Symbol(first(registered_projects))

  # load default parameters
  prms = Dict(string(s) => parameters(Val(s)) for s in default_projects)
  prms[string(project_name)] = parameters(project_name)
  # TODO parameters(Val) does not guarantee that dicts returned from parameter are flat.
  prms = Dict(k => flatten(v) for (k,v) in pairs(prms))

  # merge defaults with with user parameters
  for (sec, u_prms) in pairs(user_prms)
    def_prms = prms[sec]
    for (k,v) in pairs(u_prms)
      if !haskey(def_prms, k)
        error("Invalid parameter file: Unsupported parameter found in section [ $sec ]: '$k = $v'. Keep your parameter files clean!")
      end
      def_prms[k] = v
    end
  end

  # use a database to keep track of which parameters will be used
  prms = ParameterDB( Dict( k => ParameterDict(v) for (k,v) in pairs(prms) ) )

  return project_name, prms
end


function setup(project_name, prms, outputdir)

  mesh = Mesh(; N=prms["Mesh"]["n"], K=prms["Mesh"]["k"], range=prms["Mesh"]["range"],
                basis=prms["Mesh"]["basis"], CFL=prms["Mesh"]["cfl"])
  # cache = Cache(mesh)
  # if mesh isa Mesh2d
  #   dxdX, dydY = coorddiffs
  #   register_variables!(cache, static_variablenames=[:dxdX, :dydY])
  #   get_static_variables(cache).dxdX .= dxdX
  #   get_static_variables(cache).dydY .= dydY
  # end
  callbacks = CallbackSet()
  callbacks_substeps = CallbackSet()
  # env = Environment(mesh, cache, callbacks, callbacks_substeps)
  env = Environment(mesh, mesh.cache, callbacks, callbacks_substeps)

  # Setup the project. This involves
  # - registering variables in cache
  # - setting up initial data
  # - setting up boundary conditions
  # - setting up callbacks
  # Need to use Val(project_name) for dispatch here, because project_name is a runtime variable.
  Project = project_type(Val(project_name))
  # TODO Till I sorted out the issue with where to place bdry condition types, I have to
  # return them here and forward them below
  project, bdryconds = Project(env, prms)


  # wrap project interface methods for useage with time stepper
  function wrapped_rhs!(state_du, state_u, t)
    @timeit TO "Wrap rhs" begin
      wrap_dynamic_variables!(env.cache, state_u)
      wrap_rhs_variables!(env.cache, state_du)
    end
    # provided by project interface
    @timeit TO "rhs" begin
      rhs!(env, project, bdryconds)
    end
  end
  function wrapped_timestep!(state_u, t, n)
    @timeit TO "Wrap timestep" begin
      wrap_dynamic_variables!(env.cache, state_u)
    end
    # provided by project interface
    @timeit TO "timestep" begin
      timestep!(env, project)
    end
  end

  # Setup project independent callbacks
  # For the time being, we let the Projects freely push callbacks into env.callbacks.
  # This should make no difference as long as we 1) don't run in parallel and 2) don't allow
  # multiple projects. Might want to move that logic later from Project to here.
  # TODO Add an AliveCallback that manages a 'status' file in the output folder
  # TODO Add priority ordering to CallbackSet and callbacks.
  # TODO Drop unused callbacks from callbackset upon creation

  # Add some 'global' callbacks

  savecb = SaveCallback(Symbol.(prms["Output"]["variables"]), mesh.cache, mesh,
                        joinpath(outputdir, "output.h5"),
                        CallbackTiming(every_dt=prms["Output"]["every_dt"],
                                       every_iteration=prms["Output"]["every_iteration"]);
                        format1d=prms["Output"]["enable1d"] ? prms["Output"]["format1d"] : nothing,
                        format2d=prms["Output"]["enable2d"] ? prms["Output"]["format2d"] : nothing)

  aligned_ts = prms["Output"]["aligned_ts"]
  timealigned_savecb, wrapped_timestep! = if length(aligned_ts) > 0 && !isnothing(savecb)
    atscb = TimeAlignedSaveCallback(aligned_ts, wrapped_timestep!, savecb)
    atscb, atscb.aligned_timestep_fn
  else
    nothing, wrapped_timestep!
  end

  plotfn(u,t) = plot(env, project)
  plotcb = PlotCallback(plotfn, CallbackTiming(every_dt=prms["Plot"]["every_dt"],
              every_iteration=prms["Plot"]["every_iteration"]), pause=prms["Plot"]["pause"])

  timercb = TimerOutputCallback(TO, joinpath(outputdir, "timer.log"),
              CallbackTiming(every_dt_walltime=prms["Output"]["timer_every_dt_walltime"]))

  pcb = ProgressCallback(prms["Evolution"]["tend"],
                         prms["Output"]["progress_every_dt_walltime"], TO)

  push!(callbacks, savecb)
  push!(callbacks, timealigned_savecb)
  push!(callbacks, plotcb)
  push!(callbacks, timercb)
  push!(callbacks, pcb)

  return env, project, wrapped_rhs!, wrapped_timestep!
end


function evolve!(env, rhs_fn, timestep_fn, prms)

  @unpack tend, method = prms["Evolution"]
  @unpack cfl = prms["Mesh"]
  @unpack cache, callbacks, callbacks_substep = env

  # sanity checks
  @toggled_assert tend > 0
  @toggled_assert cfl > 0
  @toggled_assert method in [ "midpt", "rk4", "rk4_twothirds", "rk4_ralston", "ssprk3", "lserk4", "rkf10" ]

  # TODO This should be hidden behind an interface
  u0 = reduce(vcat, fields(cache.dynamic_variables))

  pblm = ODEProblem(prms["Evolution"])

  # Let's go!!!
  evolve_ERK(pblm, rhs_fn, u0, timestep_fn, tend,
             callback_fullstep = callbacks)#, callback_substep = callbacks_substep)

end
