"""
    abstract AbstractRiemannSolver 

Supertype for `RiemannSolver` interface.
Any Subtype `Solver <: RiemannSolver` must implement the following methods
- `register_variables!(cache, rsolver::Solver)`,
- `broadcast_faces!(f, rsolver::Solver, cache::Cache, mesh::Mesh)` where `f` is a 
  `@with_signature` function.
- `broadcast_boundaries!(f, rsolver::Solver, cache::Cache, mesh::Mesh)` where `f` is a 
  `@with_signature` function.

See: `ApproxRiemannSolver`
"""
abstract type AbstractRiemannSolver end


"""
    ApproxRiemannSolver <: AbstractRiemannSolver

Approximate Riemann solver interface for numerical flux computation.
Available implementations: `lax_friedrich_flux`, `central_flux`.

---

    ApproxRiemannSolver(flux, speed, equation; lhs_bc=PeriodicBC(), rhs_bc=PeriodicBC())

Assumes an `equation::SomeEquation` that implements two `@with_signature` functions
- `flux` - compute fluxes given a state,
- `speed` - compute maximum wave speed given a state.

# Example

```julia
julia> mesh = Mesh1d(); cache = Cache(mesh);

julia> rsolver = ApproxRiemannSolver(flux, speed, equation)

# register numerical fluxes as bdry variables in cache
julia> register_variables!(cache, rsolver);

# compute numerical fluxes inplace in cache
julia> broadcast_faces!(lax_friedrich_flux, rsolver, cache, mesh)
```

"""
struct ApproxRiemannSolver{T_Flux<:Function,
                           T_Speed<:Function,
                           T_Equation<:AbstractEquation,
                           N,M} <: AbstractRiemannSolver

  flux::T_Flux
  speed::T_Speed
  equation::T_Equation
  accepts::NTuple{N,Symbol}
  returns::NTuple{M,Symbol}

  function ApproxRiemannSolver(flux, speed, equation)
    # TODO Verify that accepts and returns have been registered
    @assert has_signature(flux, equation) begin
      "ApproxRiemannSolver: Require a @with_signature function for '$flux'"
    end
    @assert has_signature(speed, equation) begin
      "ApproxRiemannSolver: Require a @with_signature function for '$speed'"
    end
    _accepts_flux, returns_flux  = signature(flux, equation)
    accepts_flux = _accepts_flux[1]
    _accepts_speed, returns_speed = signature(speed, equation)
    accepts_speed = _accepts_speed[1]
    state_indices_flux = state_indices(flux, equation)
    state_indices_speed = state_indices(speed, equation)
    @assert all(accepts_flux .=== accepts_speed) && state_indices_flux == state_indices_speed begin
      "ApproxRiemannSolver: '$flux' and '$speed' must accept the same variables"
    end
    @assert length(returns_flux) == length(state_indices_flux) begin
      "ApproxRiemannSolver: '$flux' must return one value per state variable!"
    end
    @assert length(returns_speed) == 1 begin
      "ApproxRiemannSolver: '$speed' must return exaclty one value!"
    end
    rsolve_accepts = accepts_flux
    rsolve_returns = Tuple( Symbol(:nflx_, accepts_flux[i]) for i in state_indices_flux )
    return new{typeof(flux), typeof(speed), typeof(equation),
               length(rsolve_accepts), length(rsolve_returns)}(
                            flux, speed, equation, rsolve_accepts, rsolve_returns)
  end
end


struct ApproxRiemannSolver2d{T_Flux<:Function,
                             T_Speed<:Function,
                             T_Equation<:AbstractEquation,
                             N,M} <: AbstractRiemannSolver

  flux::T_Flux
  speed::T_Speed
  equation::T_Equation
  accepts::NTuple{N,Symbol}
  returns::NTuple{M,Symbol}

  function ApproxRiemannSolver2d(flux, speed, equation)
    # TODO Verify that accepts and returns have been registered
    @assert has_signature(flux, equation) begin
      "ApproxRiemannSolver2d: Require a @with_signature function for '$flux'"
    end
    @assert has_signature(speed, equation) begin
      "ApproxRiemannSolver2d: Require a @with_signature function for '$speed'"
    end
    _accepts_flux, returns_flux  = signature(flux, equation)
    accepts_flux = _accepts_flux[1]
    _accepts_speed, returns_speed = signature(speed, equation)
    accepts_speed = _accepts_speed[1]
    state_indices_flux = state_indices(flux, equation)
    state_indices_speed = state_indices(speed, equation)
    @assert all(accepts_flux .=== accepts_speed) && state_indices_flux == state_indices_speed begin
      "ApproxRiemannSolver2d: '$flux' and '$speed' must accept the same variables"
    end
    @assert length(returns_flux) == 2 * length(state_indices_flux) begin
      "ApproxRiemannSolver2d: '$flux' must return two values per state variable!"
    end
    @assert length(returns_speed) == 1 begin
      "ApproxRiemannSolver2d: '$speed' must return exaclty one value!"
    end
    rsolve_accepts = accepts_flux
    rsolve_returns = Tuple( Symbol(:nflx_, accepts_flux[i]) for i in state_indices_flux )
    return new{typeof(flux), typeof(speed), typeof(equation),
               length(rsolve_accepts), length(rsolve_returns)}(
                            flux, speed, equation, rsolve_accepts, rsolve_returns)
  end
end


# handmade definitions for @with_signature flux functions
# - accepts the same variables as rsolve.flux, rsolve.speed
# - state_indices are taken from rsolve.flux, rsolve.speed
# - returns a tuple of tuples of numerical fluxes, e.g. ( (lhs_numflx_u,), (rhs_numflx_u,) )
#   this is different from the standard @with_signature interface, but we need it to separate
#   the StructArrays belonging to the numerical fluxes on the left and right interfaces.
state_indices(f, rsolve::AbstractRiemannSolver) = state_indices(rsolve.flux, rsolve.equation)
function signature(f, rsolve::AbstractRiemannSolver)
  return rsolve.accepts, rsolve.returns
end


state(tpl, rsolve::AbstractRiemannSolver) = state(tpl, state_indices(rsolve.flux, rsolve.equation))


function register_variables!(cache, rsolver::AbstractRiemannSolver; dont_register=false)
  @unpack flux, equation = rsolver
  returns_vars = returns(lax_friedrich_flux, rsolver)
  if !dont_register
    register_variables!(cache, bdry_variablenames=returns_vars)
  end
end


#######################################################################
#             Abstractimate Riemann Solver implementations              #
#######################################################################


project(normal::NTuple{1}, flxs::NTuple{1}) = (dot(normal,flxs),)
project(normal::NTuple{2}, flxs::NTuple{2}) = (dot(normal,flxs),)
project(normal::NTuple{1}, flxs::NTuple{N}) where N = dot.(Ref(normal),flxs)
project(normal::NTuple{2}, flxs::NTuple{N,<:NTuple}) where N = dot.(Ref(normal),flxs)

maxspeed(vl::NTuple{1}, vr::NTuple{1}) = max(abs(vl[1]), abs(vr[1]))


"""
    @with_signature lax_friedrich_flux(args, rsolver::AbstractRiemannSolver)

See: `AbstractRiemannSolver`
"""
function lax_friedrich_flux(args, rsolver::AbstractRiemannSolver, normal=(1.0,))
  args_lhs, args_rhs = args
  @unpack flux, speed, equation = rsolver
  ul, ur = state(args_lhs, rsolver), state(args_rhs, rsolver)
  vl, vr = speed(args_lhs, equation), speed(args_rhs, equation)
  fl, fr = flux(args_lhs, equation), flux(args_rhs, equation)
  nfl, nfr = project(normal, fl), project(normal, fr)
  C = maxspeed(vl,vr)
  nf = @. 0.5 * (nfl + nfr) + 0.5 * C * (ur - ul)
  nf
end
has_signature(::typeof(lax_friedrich_flux), ::AbstractRiemannSolver) = true


"""
    @with_signature lax_friedrich_flux(args, rsolver::AbstractRiemannSolver)

See: `AbstractRiemannSolver`
"""
function avg_lax_friedrich_flux(args, rsolver::AbstractRiemannSolver, normal=(1.0,))
  args_lhs, args_rhs = args
  @unpack flux, speed, equation = rsolver
  ul, ur = state(args_lhs, rsolver), state(args_rhs, rsolver)
  vl, vr = speed(args_lhs, equation), speed(args_rhs, equation)
  avg_args = @. 0.5 * (args_lhs + args_rhs)
  avg_f = flux(avg_args, equation)
  avg_nf = project(normal, avg_f)
  C = maxspeed(vl,vr)
  nf = @. avg_nf + 0.5 * C * (ur - ul)
  nf
end
has_signature(::typeof(avg_lax_friedrich_flux), ::AbstractRiemannSolver) = true


"""
    @with_signature central_flux(args, rsolver::AbstractRiemannSolver)

See: `AbstractRiemannSolver`
"""
function central_flux(args, rsolver::AbstractRiemannSolver, normal=(1.0,))
  args_lhs, args_rhs = args
  ul, ur = state(args_lhs, rsolver), state(args_rhs, rsolver)
  nf = @. 0.5 * (ul + ur)
  nf
end
has_signature(::typeof(central_flux), ::AbstractRiemannSolver) = true


"""
    @with_signature mean_flux(args, rsolver::AbstractRiemannSolver)

See: `AbstractRiemannSolver`
"""
function mean_flux(args, rsolver::AbstractRiemannSolver, normal=(1.0,))
  args_lhs, args_rhs = args
  @unpack flux, equation = rsolver
  fl, fr = flux(args_lhs, equation), flux(args_rhs, equation)
  nfl, nfr = project(normal, fl), project(normal, fr)
  nf = @. 0.5 * (nfl + nfr)
  nf
end
has_signature(::typeof(mean_flux), ::AbstractRiemannSolver) = true
