const XDMF_END = """
</Domain>
</Xdmf>
"""

mutable struct XDMFOutput
    buffer::IOBuffer
    dir::String
    h5_filename::String
    mesh_filename::String
    meshcount::Int64 # number of meshes written
    tcount::Int64 # number of time steps written
end


function XDMFOutput(dir)
    isdir(dir) && throw(ArgumentError("directory '$dir' already exists"))
    mkdir(dir)

    h5_filename = joinpath(dir, "data.h5")
    mesh_filename = joinpath(dir, "mesh.xdmf")

    # create empty files
    h5open(h5_filename, "w") do file
    end
    open(mesh_filename, "w") do file
    end

    buffer = IOBuffer()

    return XDMFOutput(buffer, dir, h5_filename, mesh_filename, 0, 0)
end


function xdmfwrite_mesh(xdmf::XDMFOutput, mesh::Mesh)

    if xdmf.meshcount >= 1
        throw(ArgumentError("Mesh has already been write"))
    end

    # buffer = IOBuffer()
    write_mesh_header(xdmf.buffer)

    # Using "w" here to overwrite in case previous calls failed
    h5open(xdmf.h5_filename, "w") do file
        write_mesh(file, xdmf.buffer, mesh)
    end

    write_mesh_end(xdmf.buffer)

    # finish up xml files
    # Using "w" here to overwrite in case previous calls failed
    open(xdmf.mesh_filename, "a") do file
        write(file, take!(xdmf.buffer))
    end

    # updating count after all writes succeeded
    xdmf.meshcount += 1

end


function write_mesh_header(buffer)
    print(buffer, """
<?xml version="1.0" ?>
<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>
<Xdmf xmlns:xi="http://www.w3.org/2003/XInclude" Version="3.0">
<Domain>
<Grid GridType="Collection" CollectionType="Spatial">
""")
    return
end


function write_mesh_end(buffer)
    write(buffer, "</Grid>\n") # close grid collection
    write(buffer, XDMF_END) # close Domain and Xdmf
end


function write_mesh(h5, buffer, mesh::Mesh2d)

    @unpack x, y = get_static_variables(mesh.cache)
    z = zeros(Float64, length(x))
    itx, ity, itz = eachcell(mesh,x), eachcell(mesh,y), eachcell(mesh,z)
    for (i,(cx,cy,cz)) in enumerate(zip(itx,ity,itz))
        nx, ny, nz = size3d(size(cx))
        g = create_group(h5, "mesh$i")
        g["x"] = view(cx,:)
        g["y"] = view(cy,:)
        g["z"] = view(cz,:)
        write(buffer, """
<Grid>
<Topology TopologyType="3DSMesh" NumberOfElements="$nz $ny $nx"/>
<Geometry GeometryType="X_Y_Z">
<DataItem Name="x" Dimensions="$nz $ny $nx" Format="HDF"> data.h5:/mesh$i/x </DataItem>
<DataItem Name="y" Dimensions="$nz $ny $nx" Format="HDF"> data.h5:/mesh$i/y </DataItem>
<DataItem Name="z" Dimensions="$nz $ny $nx" Format="HDF"> data.h5:/mesh$i/z </DataItem>
</Geometry>
</Grid>
""")
    end

end


# function write_mesh(h5, buffer, mesh::Mesh3d)
#
#     x, y, z = mesh.coords
#     itx, ity, itz = eachcell(mesh,x), eachcell(mesh,y), eachcell(mesh,z)
#     for (i,(cx,cy,cz)) in enumerate(zip(itx,ity,itz,))
#         nx, ny, nz = size3d(size(cx))
#         g = create_group(h5, string(i))
#         g["x"] = cx
#         g["y"] = cy
#         g["z"] = cz
#         write(buffer, """
# <Grid>
# <Topology TopologyType="3DSMesh" NumberOfElements="$nz $ny $nx"/>
# <Geometry GeometryType="X_Y_Z">
# <DataItem Name="x" Dimensions="$nz $ny $nx" Format="HDF"> data.h5:/mesh$i/x </DataItem>
# <DataItem Name="y" Dimensions="$nz $ny $nx" Format="HDF"> data.h5:/mesh$i/y </DataItem>
# <DataItem Name="z" Dimensions="$nz $ny $nx" Format="HDF"> data.h5:/mesh$i/z </DataItem>
# </Geometry>
# </Grid>
# """)
#     end
#
# end


size3d(sz::NTuple{1}) = (sz[1],1,1)
size3d(sz::NTuple{2}) = (sz[1],sz[2],1)
# size3d(sz::NTuple{3}) = (sz[1],sz[2],sz[3])


function write_scalar(h5, buffer, mesh::Mesh, dir, tcount, t, var, name)

    write(buffer, """
<Grid GridType="Collection" CollectionType="Temporal"><Grid GridType="Collection" CollectionType="Spatial">
<Time Value="$t"/>
""")

    variter = eachcell(mesh, var)

    # a group to gather grids from one time step
    tgroup = create_group(h5, string(tcount))

    for (i,v) in enumerate(variter)
        nx, ny, nz = size3d(size(v))
        g = create_group(tgroup, string(i))
        g[name] = v
        write(buffer, """
<Grid Name="$i"><xi:include href="mesh.xdmf" xpointer="xpointer(/Xdmf/Domain/Grid[1]/Grid[$i]/child::*)"/><Attribute Name="$name"><DataItem Dimensions="$nz $ny $nx" Format="HDF"> data.h5:/$tcount/$i/$name </DataItem></Attribute></Grid>
""")
    end

    write(buffer, "</Grid></Grid>")

    filename = joinpath(dir, "$name.xdmf")
    existed = isfile(filename)
    open(filename, "a") do file
        if !existed
            print(file, """
<?xml version="1.0" ?>
<!DOCTYPE Xdmf SYSTEM "Xdmf.dtd" []>
<Xdmf xmlns:xi="http://www.w3.org/2003/XInclude" Version="3.0">
<Domain>
""")
        else
            skip(file, -length(XDMF_END))
        end
        write(file, take!(buffer))
        write(file, XDMF_END)
    end

    return
end

function xdmfwrite_scalar(xdmf::XDMFOutput, mesh::Mesh, t, vars, names)

    tcount = xdmf.tcount + 1
    h5 = h5open(xdmf.h5_filename, "r+")

    try
        for (var, name) in zip(vars, names)
            write_scalar(h5, xdmf.buffer, mesh, xdmf.dir, tcount, t, var, name)
        end
    catch e
        rethrow(e)
    finally
        close(h5)
    end

    # updating count after all writes succeeded
    xdmf.tcount = tcount

    return

end
