@testset "@with_signature" begin

  struct DummyEquation end
  dummyeq = DummyEquation()

  # general usage
  @with_signature function fn_general(eq::DummyEquation)
    @accepts u1, u2
    rhs_u1, rhs_u2 = 2.0 * u1 + u2, u2 * 3.0 - u1
    @returns rhs_u1, rhs_u2
  end
  @test accepts(fn_general, dummyeq) == ((:u1, :u2),)
  @test returns(fn_general, dummyeq) == (:rhs_u1, :rhs_u2)
  @test state_indices(fn_general, dummyeq) == ()
  @test has_signature(fn_general, dummyeq) == true
  args = (1.0, 2.0)
  obtained = fn_general(args, dummyeq)
  expected = (4.0, 5.0)
  @test obtained == expected

  # @accepts and @returns with only one argument each
  @with_signature function fn_accepts1(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test accepts(fn_accepts1, dummyeq) == ((:u1,),)
  @test returns(fn_accepts1, dummyeq) == (:rhs_u1,)
  @test state_indices(fn_accepts1, dummyeq) == ()
  @test has_signature(fn_accepts1, dummyeq) == true
  args = (1.0,)
  obtained = fn_accepts1(args, dummyeq)
  expected = (2.0,)
  @test obtained == expected

  # multiple @accepts statements
  @with_signature function fn_accepts2(eq::DummyEquation)
    @accepts u1
    @accepts v1, v2
    rhs_u1 = 2.0 * u1
    rhs_v1 = -3.0 * v1
    rhs_v2 = -3.0 * v2
    @returns rhs_u1, rhs_v1, rhs_v2
  end
  @test accepts(fn_accepts2, dummyeq) == ((:u1,),(:v1,:v2))
  @test returns(fn_accepts2, dummyeq) == (:rhs_u1,:rhs_v1,:rhs_v2)
  @test state_indices(fn_accepts2, dummyeq) == ()
  @test has_signature(fn_accepts2, dummyeq) == true
  args1 = (1.0,)
  args2 = (2.0,3.0)
  obtained = fn_accepts2(args1, args2, dummyeq)
  expected = (2.0,-6.0,-9.0)
  @test obtained == expected

  # Prefix @accepts variables
  @with_signature function fn_prefix1(eq::DummyEquation)
    @accepts Prefix(tmp), u1
    rhs_u1 = 2.0 * tmp_u1
    @returns rhs_u1
  end
  @test accepts(fn_prefix1, dummyeq) == ((:u1,),)
  @test returns(fn_prefix1, dummyeq) == (:rhs_u1,)
  @test state_indices(fn_prefix1, dummyeq) == ()
  @test has_signature(fn_prefix1, dummyeq) == true
  args = (1.0,)
  obtained = fn_prefix1(args, dummyeq)
  expected = (2.0,)
  @test obtained == expected

  # Prefix multiple @accepts variables
  @with_signature function fn_prefix2(eq::DummyEquation)
    @accepts Prefix(lhs), u1, u2
    @accepts Prefix(rhs), u1, u2
    @accepts n1, n2
    flx_u1 = n1 * lhs_u1 + n2 * rhs_u1
    flx_u2 = n1 * lhs_u2 + n2 * rhs_u2
    @returns flx_u1, flx_u2
  end
  @test accepts(fn_prefix2, dummyeq) == ((:u1,:u2),(:u1,:u2),(:n1,:n2))
  @test returns(fn_prefix2, dummyeq) == (:flx_u1,:flx_u2)
  @test state_indices(fn_prefix2, dummyeq) == ()
  @test has_signature(fn_prefix2, dummyeq) == true
  args1 = (1.0,-1.0)
  args2 = (2.0,-2.0)
  args3 = (1.0,2.0)
  obtained = fn_prefix2(args1, args2, args3, dummyeq)
  expected = (5.0,-5.0)
  @test obtained == expected

  # @accepts State variables
  @with_signature function fn_state1(eq::DummyEquation)
    @accepts State(u1, u2)
    rhs_u1, rhs_u2 = 2.0 * u1 + u2, u2 * 3.0 - u1
    @returns rhs_u1, rhs_u2
  end
  @test accepts(fn_state1, dummyeq) == ((:u1, :u2),)
  @test returns(fn_state1, dummyeq) == (:rhs_u1, :rhs_u2)
  @test state_indices(fn_state1, dummyeq) == (1,2)
  @test has_signature(fn_state1, dummyeq) == true
  args = (1.0, 2.0)
  obtained = fn_state1(args, dummyeq)
  expected = (4.0, 5.0)
  @test obtained == expected

  # @accepts single State variable
  @with_signature function fn_state2(eq::DummyEquation)
    @accepts State(u1)
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test accepts(fn_state2, dummyeq) == ((:u1,),)
  @test returns(fn_state2, dummyeq) == (:rhs_u1,)
  @test state_indices(fn_state2, dummyeq) == (1,)
  @test has_signature(fn_state2, dummyeq) == true
  args = (1.0,)
  obtained = fn_state2(args, dummyeq)
  expected = (2.0,)
  @test obtained == expected

  # @accepts mix of variables and State variables
  @with_signature function fn_state3(eq::DummyEquation)
    @accepts State(u1), u2, State(u3, u4)
    rhs_u1 = 2.0 * u1 + u2 + u3 + u4
    @returns rhs_u1
  end
  @test accepts(fn_state3, dummyeq) == ((:u1,:u2,:u3,:u4),)
  @test returns(fn_state3, dummyeq) == (:rhs_u1,)
  @test state_indices(fn_state3, dummyeq) == (1,3,4)
  @test has_signature(fn_state3, dummyeq) == true
  args = (1.0,2.0,3.0,4.0)
  obtained = fn_state3(args, dummyeq)
  expected = (11.0,)
  @test obtained == expected

  # no variable name for dispatch type required
  @with_signature function fn_noargname(#= missing name here =#::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test accepts(fn_noargname, dummyeq) == ((:u1,),)
  @test returns(fn_noargname, dummyeq) == (:rhs_u1,)
  @test state_indices(fn_noargname, dummyeq) == ()
  @test has_signature(fn_noargname, dummyeq) == true
  args = (1.0,)
  obtained = fn_noargname(args, dummyeq)
  expected = (2.0,)
  @test obtained == expected

  # function with unused variable rhs_u2 in body throws normal UndefVarError
  @with_signature function fn_undefvar(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1, rhs_u2
  end
  args = (1.0,)
  @test_throws UndefVarError fn_undefvar(args, dummyeq)

  # missing function body error
  @test_throws LoadError @eval @with_signature function fn(eq::DummyEquation) end

  # @accepts/@returns must appear at the start/end of the function body
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    println("this line is not allowed")
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
    println("this line is not allowed")
  end

  # @accepts/@returns can only appear once
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1, u2
    rhs_u1 = 2.0 * u1
    rhs_u2 = 2.0 * u2
    @returns rhs_u1, rhs_u2
    @returns rhs_u1, rhs_u2
  end

  # can't mix State and Prefix
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts Prefix(u1), State(u2)
    rhs_u1 = 2.0 * u1
    rhs_u2 = 2.0 * u2
    @returns rhs_u1, rhs_u2
  end
  # can't mix State and Prefix
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts State(u1), u2, Prefix(u3)
    rhs_u1 = 2.0 * u1
    rhs_u2 = 4.0 * u2
    rhs_u3 = 5.0 * u3
    @returns rhs_u1, rhs_u2, rhs_u3
  end

  # Prefix must appear as the first argument in a @accepts line
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1, Prefix(lhs)
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    @accepts u2, Prefix(lhs)
    rhs_u1 = 2.0 * u1
    rhs_u2 = 4.0 * u2
    @returns rhs_u1, rhs_u2
  end
  # Prefix requires at least one variable following it
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts Prefix(lhs)
    @accepts u2
    rhs_u1 = 2.0
    rhs_u2 = 4.0 * u2
    @returns rhs_u1, rhs_u2
  end

  # State can only be used with one @accepts statement
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts State(u1), u2
    @accepts u3
    rhs_u1 = 2.0 * u1
    rhs_u2 = 4.0 * u2
    rhs_u3 = 5.0 * u3
    @returns rhs_u1, rhs_u2, rhs_u3
  end

  # @accepts and @returns require each at least one argument
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts
    rhs_u1 = 1
    @returns rhs_u1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns
  end

  # function definition requires exactly one argument
  @test_throws LoadError @eval @with_signature function fn_w_error(#=no argument=#)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test_throws LoadError @eval @with_signature function fn_w_error(eq::DummyEquation, too_much)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end

  # the only argument must be type annotated
  @test_throws LoadError @eval @with_signature function fn_w_error(eq)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end

  # fallback interface methods error when called
  function fn_with_no_signature(eq::DummyEquation)
    @accepts u1
    rhs_u1 = 2.0 * u1
    @returns rhs_u1
  end
  @test_throws ErrorException accepts(fn_with_no_signature,dummyeq)
  @test_throws ErrorException returns(fn_with_no_signature,dummyeq)
  @test_throws ErrorException state_indices(fn_with_no_signature,dummyeq)
  # expect the trait function is just false
  has_signature(fn_with_no_signature, dummyeq) == false

end
